"""
Fill a worksheet with information about the prices
of the game's assets (for the duration of the game).
"""

from itertools import groupby

from openpyxl.utils import get_column_letter

from ....stocks.models import AssetPrice
from ..worksheet_filler import BaseWorksheetFiller
from .row_generator import RowGenerator

class AssetPricesWorksheetFiller(BaseWorksheetFiller):
    """
    Fill a worksheet with information about the prices
    of the game's assets (for the duration of the game).
    """

    title = "Assets"

    header_label = ('Date',)

    @property
    def header_data(self):
        """
        Implementation for the header_data property that simply returns
        the ticker strings of the assets we're reporting on.
        """

        return self.portfolio_asset_tickers

    @property
    def portfolio_asset_tickers(self):
        """The ticker strings of the assets we're reporting on."""

        return (
            asset.ticker
            for asset in self.portfolio_assets_by_pk.values()
        )

    @property
    def asset_prices(self):
        """
        An iterator over asset prices of all assets we're reporting
        on, sorted by date.
        """

        game = self.game

        return AssetPrice.objects.filter(
            asset__in=self.portfolio_assets_by_pk.keys(),
            closing_date__gte=game.start_date,
            closing_date__lte=game.end_date,
            closing_price__isnull=False,
        ).order_by('closing_date')

    @property
    def grouped_asset_prices(self):
        """
        An iterator over asset prices of all assets we're reporting
        on, grouped by date. See `itertools.groupby()`.
        """

        return groupby(
            self.asset_prices,
            key=lambda asset_price: asset_price.closing_date,
        )

    def add_data(self):
        """
        Implementation of `add_data` that adds price data for all assets
        on all dates.
        """

        for today, todays_asset_prices in self.grouped_asset_prices:
            self.add_todays_price_data(today, todays_asset_prices)

    def add_todays_price_data(self, today, todays_asset_prices):
        """
        Implementation of `add_data` that adds price data for all assets
        on one date.
        """

        values = self.combine_todays_price_data(today, todays_asset_prices)
        self.append_row(*values)

    def combine_todays_price_data(self, today, todays_asset_prices):
        """Combine the label and price data for one day."""

        return [
            *self.collect_todays_price_label(today),
            *self.collect_todays_price_data(todays_asset_prices),
        ]

    def collect_todays_price_label(self, today):
        """Generate the label value for one day."""

        return [today]

    def collect_todays_price_data(self, todays_asset_prices):
        """
        Return an iterable containing price values in the order
        of the header row, with None values if one asset does not
        have a price value in todays_asset_prices.
        """

        return RowGenerator(self, todays_asset_prices).values

    def set_all_columns_to_autofit(self):
        """Helper function that sets all column widths to 'autofit'."""

        worksheet = self.worksheet
        for column in range(1, worksheet.max_column + 1):
            self.set_column_autofit(column)

    def set_column_autofit(self, column):
        """Helper function that sets one column's width to 'autofit'."""

        column_letter = get_column_letter(column)
        self.worksheet.column_dimensions[column_letter].autoFit = True

    def finalize(self):
        """Perform the final touch-ups after all data is added to the sheet."""

        self.set_all_columns_to_autofit()
        self.freeze_panes('B2')

    @property
    def portfolio_assets_by_pk(self):
        """
        Provide easy access to the portfolio assets, indexed by pk
        (which is a property of the generator).
        """

        return self.generator.portfolio_assets_by_pk
