"""
Generate an Excel spreadsheet (represented as a bytes object)
suitable for a download by an admin.
"""

from functools import cached_property
from copy import copy
from itertools import chain

from openpyxl import Workbook as ExcelWorkbook
from openpyxl.writer.excel import save_virtual_workbook
from openpyxl.styles import Alignment, DEFAULT_FONT

from .portfolio_asset_weights import PortfolioAssetWeightsWorksheetFiller
from .revision_results import RevisionResultsWorksheetFiller
from .asset_prices import AssetPricesWorksheetFiller

class DownloadGenerator:
    """
    Generate an Excel spreadsheet (represented as a bytes object)
    suitable for a download by an admin.
    """

    def __init__(self, game):
        self.game = game

    @cached_property
    def result(self):
        """
        A spreadsheet containing information suitable for download,
        as a bytes object.
        """

        self.add_worksheets()
        return save_virtual_workbook(self.workbook)

    @cached_property
    def worksheet_generator(self):
        """A generator that yields worksheets to use."""

        workbook = self.workbook
        yield workbook.active
        while True:
            yield workbook.create_sheet()

    def get_next_worksheet(self):
        """Return the next worksheet to use, creating it as neccessary."""

        return next(self.worksheet_generator)

    # An iterable of worksheet fillers. Each will be used in turn to fill
    # a worksheet. Using a tuple as an ersatz immutable list.
    worksheet_fillers = (
        PortfolioAssetWeightsWorksheetFiller,
        RevisionResultsWorksheetFiller,
        AssetPricesWorksheetFiller,
    )

    def add_worksheets(self):
        """Add worksheets according to the worksheet_fillers property."""

        for worksheet_filler in self.worksheet_fillers:
            worksheet = self.get_next_worksheet()
            worksheet_filler(self, worksheet).fill()

    @cached_property
    def header_font(self):
        """An openpyxl Font object suitable for headers."""

        header_font = copy(DEFAULT_FONT)
        header_font.bold = True
        return header_font

    @cached_property
    def assets(self):
        """All Asset objects belonging to the game, sorted."""

        return list(self.game.assets.order_by('ticker'))

    @cached_property
    def participations(self):
        """All participations belonging to the game, sorted."""

        return self.game.participation_set.select_related(
            'user',
        ).prefetch_related(
            'portfolio_set',
            'portfolio_set__portfoliorevision_set',
            'portfolio_set__portfoliorevision_set__portfolioasset_set',
            'portfolio_set__portfoliorevision_set__portfolioasset_set__asset',
        ).order_by(
            'user__full_name',
            'pk',
        )

    @cached_property
    def workbook(self):
        """The openpyxl workbook object that will contain all information."""

        workbook = ExcelWorkbook()
        workbook.iso_dates = True
        return workbook

    @cached_property
    def default_alignment(self):
        """The openpyxl Alignment object that will be applied to all cells."""

        return Alignment(vertical='center')

    @cached_property
    def revisions_by_pk(self):
        """
        All revisions of all portfolios of all participants of the game,
        as a dict indexed by the pk of each revision.
        """

        return {
            revision.pk: revision
            for revision in self.revisions
        }

    @property
    def revisions(self):
        """An iterable of all portfolios of all participants of the game."""

        return chain.from_iterable(
            self.sort_revisions(portfolio.portfoliorevision_set.all())
            for portfolio in self.portfolios
        )

    @property
    def portfolios(self):
        """
        An iterable of all portfolios of all portfolios of all participants
        of the game.
        """

        return chain.from_iterable(
            self.sort_portfolios(participation.portfolio_set.all())
            for participation in self.participations
        )

    @cached_property
    def portfolio_assets_by_pk(self):
        """
        All portfolio asset objects referenced by any of the
        portfolios, as a dict indexed by the asset they refer to.
        """

        return {
            asset.pk: asset
            for asset in self.sorted_portfolio_assets
        }

    @property
    def sorted_portfolio_assets(self):
        """
        A sorted list of portfolio asset objects referenced by any of the
        portfolios. Is likely to contain duplicate items.
        """

        return sorted(
            self.portfolio_assets,
            key=lambda asset: (asset.ticker, asset.pk),
        )

    @property
    def portfolio_assets(self):
        """
        An iterable of asset objects referenced by any of the
        portfolios. Is likely to contain duplicate items.
        """

        return (
            asset_weight.asset
            for asset_weight in self.asset_weights
        )

    @property
    def asset_weights(self):
        """
        All portfolio asset objects of all revisions of all
        portfolios of all participants of the game.
        """

        return chain.from_iterable(
            revision.portfolioasset_set.all()
            for revision in self.revisions
        )

    def sort_portfolios(self, portfolios):
        """Return the given portfolios in a new, sorted list."""

        return sorted(
            portfolios,
            key=lambda portfolio: (
                portfolio.name.lower(),
                portfolio.name,
                portfolio.pk,
            ),
        )

    def sort_revisions(self, revisions):
        """Return the given revisions in a new, sorted list."""

        return sorted(
            revisions,
            key=lambda revision: revision.date,
        )
