"""Base class for portfolio revision based worksheet fillers"""

from abc import abstractmethod

from .worksheet_filler import BaseWorksheetFiller

class RevisionWorksheetFiller(BaseWorksheetFiller):
    """Base class for portfolio revision based worksheet fillers"""

    # Header strings for the label part, using a tuple as an ersatz
    # immutable list.
    header_label = (
        'Student',
        'SNR',
        'Portfolio',
        'Revision',
    )

    def add_data(self):
        """Fill the data (non-header) part of the worksheet."""

        for participation in self.participations:
            self.add_participation_data(participation)

    def add_participation_data(self, participation):
        """Add rows for one participation."""

        participation_start_row = self.row_count + 1
        portfolios = self.sort_portfolios(participation.portfolio_set.all())
        self.add_portfolios_data(portfolios)
        self.group_participation_data(participation_start_row)

    def add_portfolios_data(self, portfolios):
        """Add rows for each portfolio of a participation."""

        for portfolio in portfolios:
            self.add_portfolio_data(portfolio)

    def add_portfolio_data(self, portfolio):
        """Add rows for one portfolio of a participation."""

        portfolio_start_row = self.row_count + 1
        revisions = self.sort_revisions(portfolio.portfoliorevision_set.all())
        self.add_revisions_data(revisions)
        self.group_portfolio_data(portfolio_start_row)

    def add_revisions_data(self, revisions):
        """Add rows for all revisions of one portfolio of a participation."""

        for revision in revisions:
            self.add_revision_data(revision)

    def add_revision_data(self, revision):
        """Add a row for one revision of one portfolio of a participation."""

        row_data = self.combine_revision_data(revision)
        self.append_row(*row_data)

    def combine_revision_data(self, revision):
        """Return the combined label and data for one revision."""

        return [
            *self.collect_revision_label(revision),
            *self.collect_revision_data(revision),
        ]

    @abstractmethod
    def collect_revision_data(self, revision):
        """Return data for one revision. To be implemented by a subclass."""

    def collect_revision_label(self, revision):
        """Return the label for one revision."""

        portfolio = revision.portfolio
        participation = portfolio.participation
        user = participation.user
        return [
            user.full_name,
            user.student_number,
            portfolio.name,
            revision.date,
        ]

    def group_portfolio_data(self, portfolio_start_row):
        """Group the data for a portfolio by merging cells."""

        self.worksheet.merge_cells(
            start_row=portfolio_start_row,
            start_column=3,
            end_row=self.row_count,
            end_column=3,
        )

    def group_participation_data(self, participation_start_row):
        """Group the data for a participation by merging cells."""

        participation_end_row = self.row_count
        worksheet = self.worksheet
        worksheet.merge_cells(
            start_row=participation_start_row,
            start_column=1,
            end_row=participation_end_row,
            end_column=1,
        )
        worksheet.merge_cells(
            start_row=participation_start_row,
            start_column=2,
            end_row=participation_end_row,
            end_column=2,
        )

    def finalize(self):
        """Apply final touch-ups to the worksheet after all data is added."""

        self.set_column_width('A', 20)
        self.set_column_width('B', 10)
        self.set_column_width('C', 20)
        self.set_column_width('D', 10)
        self.freeze_panes('D2')

    @property
    def participations(self):
        """
        Provide easy access to the participations of the game
        (which is a property of the generator).
        """

        return self.generator.participations

    @property
    def sort_portfolios(self):
        """
        Provide easy access to the sort_portfolios method of the generator.
        """

        return self.generator.sort_portfolios

    @property
    def sort_revisions(self):
        """
        Provide easy access to the sort_revisions method of the generator.
        """

        return self.generator.sort_revisions
