"""
Fill a worksheet with information about the sharpe ratios of
each revision of each portfolio of each participant, for each
date of the duration of the game.
"""

from itertools import chain
from functools import cached_property

from ...compute import compute_revision_results
from ..revision import RevisionWorksheetFiller
from .row_generator import RevisionResultsRowGenerator

class RevisionResultsWorksheetFiller(RevisionWorksheetFiller):
    """
    Fill a worksheet with information about the sharpe ratios of
    each revision of each portfolio of each participant, for each
    date of the duration of the game.
    """

    title = "Results"

    def collect_revision_data(self, revision):
        """
        Return values suitable for use in a row of results data for one
        portfolio revision.
        """

        revision_results = self.revisions_results.get(revision.pk, [])
        return RevisionResultsRowGenerator(self, revision_results).values

    @property
    def header_data(self):
        """
        The values used in the header of this worksheet: the dates for
        which at least one of the portfolio revisions has a known result.
        """
        return self.results_dates

    @cached_property
    def results_dates(self):
        """
        A sorted list of dates for which at least one of the portfolio
        revisions has a known result.
        """

        return sorted(self.unique_result_dates)

    @property
    def unique_result_dates(self):
        """
        The unique dates for which at least one of the portfolio
        revisions has a known result.
        """

        return set(self.concatenated_and_filtered_result_dates)

    @property
    def concatenated_and_filtered_result_dates(self):
        """
        An iterable of dates for which at least one of the portfolio
        revisions has a known result. Contains duplicate entries.
        """

        return (
            result['date']
            for result in self.concatenated_results
            if 'sharpe_ratio' in result
        )

    @property
    def concatenated_results(self):
        """
        An iterable of all results of all revisions of all portfolios
        of all participants of the game.
        """

        return chain.from_iterable(self.revisions_results.values())

    @cached_property
    def revisions_results(self):
        """The computed results for this game."""

        game = self.game
        return compute_revision_results(
            self.revisions_by_pk.values(),
            start_date=game.start_date,
            end_date=game.end_date,
            risk_free_return=game.risk_free_return,
        )

    @property
    def revisions_by_pk(self):
        """
        Provide easy access to the portfolio revisions, indexed by pk
        (which is a property of the generator).
        """

        return self.generator.revisions_by_pk

