"""Base class for worksheet fillers."""

from abc import ABC, abstractmethod

class BaseWorksheetFiller(ABC):
    """
    Base class for worksheet fillers.

    Worksheet fillers fill the supplied openpyxl worksheet with information.

    Usage: instantiate a subclass and call the fill() method.
    """

    def __init__(self, generator, worksheet):
        self.generator = generator
        self.worksheet = worksheet

    def fill(self):
        """Main entry point for this worksheet filler."""

        self.set_title()
        self.add_header()
        self.add_data()
        self.finalize()

    def set_title(self):
        """Set the worksheet title from `self.title`."""

        self.worksheet.title = self.title

    @property
    @abstractmethod
    def title(self):
        """The title of this worksheet. To be implemented in a subclass."""

    def add_header(self):
        """Utility function to add a row that is styled as a header."""

        self.append_row(*self.header_combined, is_header=True)

    @property
    def header_combined(self):
        """The header label (left-hand side) combined with the header data."""

        return [
            *self.header_label,
            *self.header_data,
        ]

    header_label = ()
    header_data = ()

    @abstractmethod
    def add_data(self):
        """
        Called when it's time to add data to the sheet.
        Override in subclass.
        """

    def finalize(self):
        """
        Called to add any finishing touches to the worksheet,
        such as freezing panes. Override in subclass if desired.
        """

    @property
    def row_count(self):
        """
        Number of rows in the spreadsheet.

        max_row equals 1 both when the worksheet is empty and when
        it contains exactly one row. This makes it unreliable as a
        tool to find the first empty row. This function always returns
        the correct number of rows in the worksheet.
        """

        worksheet = self.worksheet

        for _ in worksheet.iter_rows():
            # We only get here if there's at least one row:
            return worksheet.max_row

        # There were no rows:
        return 0

    def append_row(self, *values, is_header=None):
        """
        Append an iterable of values as a row.
        A `None` value causes that column to be skipped.
        If `is_header` is true, header-specific styling is applied.
        """

        row_number = self.row_count + 1
        for index, value in enumerate(values, start=1):
            self.add_cell(value, row_number, index, is_header)

    def add_cell(self, value, row, column, is_header):
        """
        If value is `None`, do nothing.
        Otherwise, add a spreedsheet cell with the given coordinates
        and value.
        If `is_header` is true, header-specific styling is applied.
        """

        if value is not None:
            self.create_cell(value, row, column, is_header)

    def create_cell(self, value, row, column, is_header):
        """
        Add and format a spreedsheet cell with the given coordinates and
        value.
        If `is_header` is true, header-specific styling is applied.
        """
        cell = self.worksheet.cell(row=row, column=column, value=value)
        self.format_cell(cell, is_header)

    def format_cell(self, cell, is_header):
        """
        Format the given spreedsheet cell.
        If `is_header` is true, header-specific styling is applied.
        """
        cell.alignment = self.default_alignment
        if is_header:
            cell.font = self.header_font

    def set_column_width(self, column, width):
        """Set the width of the given column."""

        self.worksheet.column_dimensions[column].width = width

    def freeze_panes(self, cell_identification_string):
        """
        Set the scrollable area of the worksheet.
        The cell identification string indicates the top-leftmost cell
        that is still scrollable. Everything above and to the left of
        that cell will remain visible in the viewport when the worksheet
        content is scrolled.
        It is formatted as a spreadsheet coordinate string, such as 'B2'.
        """

        worksheet = self.worksheet
        worksheet.freeze_panes = worksheet[cell_identification_string]

    @property
    def default_alignment(self):
        """
        Provide easy access to the default alignment style (which is a
        property of the generator).
        """

        return self.generator.default_alignment

    @property
    def header_font(self):
        """
        Provide easy access to the header font style (which is a
        property of the generator).
        """

        return self.generator.header_font

    @property
    def assets(self):
        """
        Provide easy access to the list of assets for this game (which is a
        property of the generator).
        """

        return self.generator.assets

    @property
    def game(self):
        """
        Provide easy access to the game (which is a property of the
        generator).
        """

        return self.generator.game
