"""UvT app related tests"""

from json import loads as parse_json
from django.test import TestCase
from rest_framework.test import APIClient
from investment_game.uvt.models import User


class TestUsers(TestCase):
    """Test user-related API endpoints"""

    fixtures = ['stocks', 'players', 'game']
    client_class = APIClient

    def test_admin_list(self):
        """An admin should be able to fetch everyone's data"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/users')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 35)

    def test_admin_get(self):
        """An admin should be able to fetch anyone's data"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/users/1')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(content['username'], 'schmeits')

    def test_player_get(self):
        """A player should be able to fetch their own data"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/users/8')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(content['username'], 'u000901')

    def test_player_get_other(self):
        """A player should not be able to fetch other people's data"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/users/1')
        self.assertEqual(
            response.status_code,
            404,
            "player should not be able to fetch other people's data",
        )

    def test_player_create(self):
        """Creating accounts should not be possible."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.post('/users', {
            'username': 'bob',
            'is_admin': False,
        }, format='json')

        self.assertEqual(response.status_code, 403)

    def test_admin_create(self):
        """Creating accounts should not be possible."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/users', {
            'username': 'bob',
            'is_admin': False,
        }, format='json')

        self.assertEqual(response.status_code, 403)

    def test_admin_update(self):
        """Admins may modify accounts."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/users/30')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)

        content['is_admin'] = True
        response = client.put('/users/30', content, format='json')
        self.assertEqual(response.status_code, 200, response.content)

    def test_player_update(self):
        """Players must not be able to modify accounts."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/users/8')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)

        content['is_admin'] = True
        response = client.put('/users/8', content, format='json')
        self.assertEqual(response.status_code, 403)

        response = client.patch('/users/8', content, format='json')
        self.assertEqual(response.status_code, 403)

    def test_player_delete(self):
        """Deleting accounts should not be possible."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.delete('/users/8')
        self.assertEqual(response.status_code, 403)

    def test_admin_delete(self):
        """Deleting accounts should not be possible."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.delete('/users/8')
        self.assertEqual(response.status_code, 403)

    def test_admin_get_participations(self):
        """An admin should be able to fetch anyone's list of participations"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/users/9/participations')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 3)

    def test_player_get_participations(self):
        """A player should be able to fetch their list of participations"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/users/9/participations')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 2)

    def test_player_get_participations_other(self):
        """
        A player should not be able to fetch other people's participations
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/users/6/participations')
        self.assertEqual(response.status_code, 403, response.content)

    def test_admin_get_games(self):
        """An admin should be able to fetch anyone's list of games"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/users/9/games')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 3)

    def test_player_get_games(self):
        """A player should be able to fetch their list of games"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/users/9/games')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 2)

    def test_player_get_games_other(self):
        """
        A player should not be able to fetch other people's games
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/users/6/games')
        self.assertEqual(response.status_code, 403, response.content)

    def test_admin_get_games_participations(self):
        """
        An admin should be able to fetch the participation for anyone's
        game
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/users/9/games/1/participations')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 1)

    def test_player_get_games_participations(self):
        """A player should be able to fetch the participation for their game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/users/9/games/1/participations')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertEqual(len(content), 1)

    def test_player_get_games_participations_other(self):
        """
        A player should not be able to fetch the participation for other
        people's games
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/users/6/games/1/participations')
        self.assertEqual(response.status_code, 403, response.content)
